from tests.fixtures import mainnet_only
from y.prices import magic


@mainnet_only
def test_stargate_lp_price():
    token = "0xdf0770dF86a8034b3EFEf0A1Bb3c889B8332FF56"  # S*USDC
    price = magic.get_price(token, 17_000_000, skip_cache=True)
    assert price and price > 0



from functools import lru_cache

from dank_mids.types import _DictStruct
from inflection import underscore


class _CamelDictStruct(_DictStruct, rename="camel"):
    """
    A mixin class that allows camelCase key lookups for snake_case struct attrs
    Original use case was so Log structs can be used interchangably with LogReceipt instances
    """

    def __getitem__(self, attr: str):
        return getattr(self, _make_snake(attr))


class Log(_CamelDictStruct):
    removed: bool | None
    log_index: int | None
    transaction_index: int | None
    transaction_hash: str
    block_hash: str | None
    block_number: int | None
    address: str | None
    data: str | None
    topics: list[str] | None


class Trace(_CamelDictStruct):
    # TODO so we can trace chain for eth_port on alt chains
    pass


@lru_cache(maxsize=None)
def _make_snake(camel: str) -> str:
    return underscore(camel)



from asyncio import Task, create_task
from collections.abc import AsyncIterator, Awaitable, Callable
from contextlib import suppress
from enum import IntEnum
from logging import DEBUG, getLogger
from typing import Any, NewType, TypeVar

import a_sync
from a_sync import cgather
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import ZERO_ADDRESS
from brownie.convert.datatypes import EthAddress
from brownie.network.event import _EventItem
from eth_abi.exceptions import InvalidPointer
from hexbytes import HexBytes
from multicall import Call
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import constants, contracts
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20, ContractBase, WeiBalance
from y.constants import CHAINID, CONNECTED_TO_MAINNET
from y.contracts import Contract
from y.datatypes import Address, AnyAddressType, Block, Pool, UsdPrice, UsdValue
from y.exceptions import ContractNotVerified, TokenNotFound
from y.networks import Network
from y.prices.dex.balancer._abc import BalancerABC, BalancerPool
from y.utils.cache import a_sync_ttl_cache
from y.utils.events import ProcessedEvents
from y.utils.logging import get_price_logger

BALANCER_V2_VAULTS = {
    Network.Mainnet: [
        "0xBA12222222228d8Ba445958a75a0704d566BF2C8",
    ],
    Network.Fantom: [
        "0x20dd72Ed959b6147912C2e529F0a0C651c33c9ce",
    ],
    Network.Polygon: [
        "0xBA12222222228d8Ba445958a75a0704d566BF2C8",
    ],
    Network.Arbitrum: [
        "0xBA12222222228d8Ba445958a75a0704d566BF2C8",
    ],
    Network.Base: [
        "0xBA12222222228d8Ba445958a75a0704d566BF2C8",
    ],
    Network.Berachain: [
        "0x4Be03f781C497A489E3cB0287833452cA9B9E80B",  # BEX Exchange https://docs.bex.berachain.com/developers/
    ],
}.get(CHAINID, [])

MESSED_UP_POOLS = {
    Network.Mainnet: [
        # NOTE: this was the first ever balancer "pool" and isn't actually a pool
        "0xF3799CBAe9c028c1A590F4463dFF039b8F4986BE",
    ],
}.get(CHAINID, [])

T = TypeVar("T")

PoolId = NewType("PoolId", bytes)
PoolBalances = dict[ERC20, WeiBalance]

logger = getLogger(__name__)


class PoolSpecialization(IntEnum):
    ComposableStablePool = 0
    WeightedPool = 1
    WeightedPool2Tokens = 2
    # This is a weird one
    CronV1Pool = -1

    @staticmethod
    def with_immutable_tokens() -> list["PoolSpecialization"]:
        """
        Get a list of pool specializations with immutable tokens.

        Returns:
            A list of :class:`~PoolSpecialization` enums representing pools with immutable tokens.

        Examples:
            >>> PoolSpecialization.with_immutable_tokens()
            [<PoolSpecialization.ComposableStablePool: 0>, <PoolSpecialization.WeightedPool: 1>, <PoolSpecialization.WeightedPool2Tokens: 2>, <PoolSpecialization.CronV1Pool: -1>]
        """
        return [
            PoolSpecialization.ComposableStablePool,
            PoolSpecialization.WeightedPool,
            PoolSpecialization.WeightedPool2Tokens,
            PoolSpecialization.CronV1Pool,
        ]


class BalancerV2Vault(ContractBase):
    def __init__(self, address: AnyAddressType, *, asynchronous: bool = False) -> None:
        """
        Initialize a BalancerV2Vault instance.

        Args:
            address: The address of the Balancer V2 Vault.
            asynchronous: Whether to use asynchronous operations.

        Examples:
            >>> vault = BalancerV2Vault("0xBA12222222228d8Ba445958a75a0704d566BF2C8")
        """
        super().__init__(address, asynchronous=asynchronous)
        self._events = BalancerEvents(
            self,
            addresses=address,
            topics=["0x3c13bc30b8e878c53fd2a36b679409c073afd75950be43d8858768e956fbc20e"],
        )
        if not self._is_cached:
            # we need the contract cached so we can decode logs correctly
            self.contract

    @stuck_coro_debugger
    async def pools(self, block: Block | None = None) -> AsyncIterator["BalancerV2Pool"]:
        """
        Asynchronously iterate over Balancer V2 pools.

        Args:
            block: The block number to query. Defaults to the latest block.

        Yields:
            Instances of :class:`~BalancerV2Pool`.

        Examples:
            >>> async for pool in vault.pools():
            ...     print(pool)
        """
        async for pool in self._events.events(to_block=block):
            yield pool

    @stuck_coro_debugger
    async def pools_for_token(
        self, token: Address, block: Block | None = None
    ) -> AsyncIterator["BalancerV2Pool"]:
        """
        Asynchronously iterate over Balancer V2 pools containing a specific token.

        Args:
            token: The address of the token to search for.
            block: The block number to query. Defaults to the latest block.

        Yields:
            Instances of :class:`~BalancerV2Pool` containing the specified token.

        Examples:
            >>> async for pool in vault.pools_for_token("0xTokenAddress"):
            ...     print(pool)
        """
        tasks = a_sync.map(BalancerV2Pool.tokens, block=block)
        debug_logs = logger.isEnabledFor(DEBUG)
        async for pool in self.pools(block=block):
            if tokens := pool._tokens:
                if token in tokens:
                    if debug_logs:
                        logger._log(DEBUG, "%s contains %s", (pool, token))
                    yield pool
            else:
                # start the task now, we can await it later
                tasks[pool]

        if tasks:
            async for pool, tokens in tasks.map(pop=True):
                if token in tokens:
                    if debug_logs:
                        logger._log(DEBUG, "%s contains %s", (pool, token))
                    yield pool

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def get_pool_tokens(self, pool_id: HexBytes, block: Block | None = None):
        """
        Get the tokens and balances for a specific pool.

        Args:
            pool_id: The ID of the pool.
            block: The block number to query. Defaults to the latest block.

        Returns:
            A tuple containing the tokens and their balances.

        Examples:
            >>> tokens, balances = await vault.get_pool_tokens(pool_id)
        """
        contract = await contracts.Contract.coroutine(self.address)
        try:
            return await contract.getPoolTokens.coroutine(pool_id, block_identifier=block)
        except InvalidPointer:
            return (), (), None

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def get_pool_info(
        self, poolids: tuple[HexBytes, ...], block: Block | None = None
    ) -> list[tuple]:
        """
        Get information for multiple pools.

        Args:
            poolids: A tuple of pool IDs.
            block: The block number to query. Defaults to the latest block.

        Returns:
            A list of tuples containing pool information.

        Examples:
            >>> pool_info = await vault.get_pool_info((pool_id1, pool_id2))
        """
        contract = await contracts.Contract.coroutine(self.address)
        return await contract.getPoolTokens.map(poolids, block_identifier=block)

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def deepest_pool_for(
        self, token_address: Address, block: Block | None = None
    ) -> "BalancerV2Pool":
        """
        Find the deepest pool for a specific token.

        Args:
            token_address: The address of the token.
            block: The block number to query. Defaults to the latest block.

        Returns:
            The :class:`~BalancerV2Pool` with the deepest liquidity for the specified token.

        Examples:
            >>> deepest_pool = await vault.deepest_pool_for("0xTokenAddress")
        """
        balance_tasks: a_sync.TaskMapping[BalancerV2Pool, WeiBalance | None]

        logger = get_price_logger(token_address, block, extra="balancer.v2")

        balance_tasks = BalancerV2Pool.get_balance.map(token_address=token_address, block=block)
        balances_aiterator = balance_tasks.map(
            self.pools_for_token(token_address, block=block), pop=True
        )
        filtered = balances_aiterator.filter(_lookup_balance_from_tuple)
        high_to_low = filtered.sort(key=_lookup_balance_from_tuple, reverse=True)

        if logger.isEnabledFor(DEBUG):
            async for pool, balance in high_to_low:
                logger._log(DEBUG, "deepest pool %s balance %s", (pool, balance))
                return pool
        else:
            async for pool, balance in high_to_low:
                return pool


class BalancerEvents(ProcessedEvents[tuple[HexBytes, EthAddress, Block]]):
    __slots__ = ("asynchronous",)

    def __init__(self, vault: BalancerV2Vault, *args, asynchronous: bool = False, **kwargs):
        """
        Initialize a BalancerEvents instance.

        Args:
            vault: The associated :class:`~BalancerV2Vault`.
            asynchronous: Whether to use asynchronous operations.

        Examples:
            >>> events = BalancerEvents(vault)
        """
        super().__init__(*args, **kwargs)
        self.vault = vault
        self.asynchronous = asynchronous
        self.__tasks: dict[Task[Any], BalancerV2Pool] = {}

    def _include_event(self, event: _EventItem) -> Awaitable[bool]:
        """
        Determine whether to include a specific event.

        Args:
            event: The event to evaluate.

        Returns:
            A boolean indicating whether to include the event.

        Examples:
            >>> include = await events._include_event(event)
        """
        if event["poolAddress"] in MESSED_UP_POOLS:
            return False
        # NOTE: For some reason the Balancer fork on Fantom lists "0x3e522051A9B1958Aa1e828AC24Afba4a551DF37d"
        #       as a pool, but it is not a contract. This handler will prevent it and future cases from causing problems.
        # NOTE: this isn't really optimized as it still runs semi-synchronously but its better than what was had previously
        return self.executor.run(contracts.is_contract, event["poolAddress"])

    def _process_event(self, event: _EventItem) -> "BalancerV2Pool":
        """
        Process a specific event and return the associated Balancer V2 pool.

        Args:
            event: The event to process.

        Returns:
            The associated :class:`~BalancerV2Pool`.

        Examples:
            >>> pool = events._process_event(event)
        """
        try:
            specialization = PoolSpecialization(event["specialization"])
        except ValueError:
            specialization = None
        pool = BalancerV2Pool(
            address=event["poolAddress"],
            id=HexBytes(event["poolId"]),
            specialization=specialization,
            vault=self.vault,
            _deploy_block=event.block_number,
            asynchronous=self.asynchronous,
        )
        # lets get this cached into memory now
        task = create_task(pool.tokens(sync=False))
        self.__tasks[task] = pool
        task.add_done_callback(self._task_done_callback)
        return pool

    def _get_block_for_obj(self, pool: "BalancerV2Pool") -> int:
        """
        Get the block number for a specific pool.

        Args:
            pool: The :class:`~BalancerV2Pool` to evaluate.

        Returns:
            The block number.

        Examples:
            >>> block_number = events._get_block_for_obj(pool)
        """
        return pool._deploy_block

    def _task_done_callback(self, t: Task):
        """
        Log any Exception that occurs in a startup task.

        Because these tasks are only used to populate caches
        and don't block anything, we don't need to propagate
        the Exception and can just remove the task so it can
        be garbage collected

        Args:
            t: The completed task.

        Examples:
            >>> events._task_done_callback(task)
        """
        pool = self.__tasks.pop(t)
        if t.cancelled():
            logger.error("balancer v2 startup task %s is cancelled??", t)
        elif exc := t.exception():
            logger.error(f"balancer v2 startup task failed: {exc!r}")
        # force shrink the dict's hash table as the tasks complete
        self.__tasks = dict(self.__tasks)


class BalancerV2Pool(BalancerPool):
    """A pool from Balancer Protocol v2"""

    # internal variables to save calls in some instances
    # they do not necessarily reflect real life at all times
    # defaults are stored as class vars to keep instance dicts smaller
    _tokens: tuple[ERC20, ...] = None
    __nonweighted: bool = False
    __weights: list[int] = None

    def __init__(
        self,
        address: AnyAddressType,
        *,
        id: HexBytes | None = None,
        specialization: PoolSpecialization | None = None,
        vault: BalancerV2Vault | None = None,
        asynchronous: bool = False,
        _deploy_block: Block | None = None,
    ):
        """
        Initialize a BalancerV2Pool instance.

        Args:
            address: The address of the pool.
            id: The ID of the pool.
            specialization: The specialization of the pool.
            vault: The associated :class:`~BalancerV2Vault`.
            asynchronous: Whether to use asynchronous operations.
            _deploy_block: The block number when the pool was deployed.

        Examples:
            >>> pool = BalancerV2Pool("0xPoolAddress")
        """
        super().__init__(address, asynchronous=asynchronous, _deploy_block=_deploy_block)
        if id is not None:
            self.id = id
        if specialization is not None:
            self.pool_type = specialization
        if vault is not None:
            self.vault = vault

    @a_sync.aka.cached_property
    async def id(self) -> PoolId:
        """
        Get the ID of the pool.

        Returns:
            The pool ID.

        Examples:
            >>> pool_id = await pool.id
        """
        return await Call(self.address, "getPoolId()(bytes32)")

    __id__: HiddenMethodDescriptor[Self, PoolId]

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def vault(self) -> BalancerV2Vault | None:
        """
        Get the associated Balancer V2 Vault.

        Returns:
            The associated :class:`~BalancerV2Vault`, or None if not found.

        Examples:
            >>> vault = await pool.vault
        """
        with suppress(ContractLogicError):
            vault = await Call(self.address, "getVault()(address)")
            if vault == ZERO_ADDRESS:
                return None
            elif vault:
                return BalancerV2Vault(vault, asynchronous=True)
        # in earlier web3 versions, we would get `None`. More recently, we get ContractLogicError. This handles both
        if CONNECTED_TO_MAINNET and await self.__build_name__ == "CronV1Pool":
            # NOTE: these `CronV1Pool` tokens ARE balancer pools but don't match the expected pool abi?
            return BalancerV2Vault("0xBA12222222228d8Ba445958a75a0704d566BF2C8", asynchronous=True)

    __vault__: HiddenMethodDescriptor[Self, BalancerV2Vault | None]

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def pool_type(self) -> PoolSpecialization | int:
        """
        Get the type of the pool.

        Returns:
            The pool type as a :class:`~PoolSpecialization` or an integer.

        Examples:
            >>> pool_type = await pool.pool_type
        """
        vault = await self.__vault__
        if vault is None:
            raise ValueError(f"{self} has no vault") from None
        elif poolid := await self.__id__:
            _, specialization = await vault.contract.getPool.coroutine(poolid)
        elif CONNECTED_TO_MAINNET and await self.__build_name__ == "CronV1Pool":
            # NOTE: these `CronV1Pool` tokens ARE balancer pools but don't match the expected pool abi?
            return PoolSpecialization.CronV1Pool
        else:
            raise ValueError(f"{self} has no poolid") from None
        try:
            return PoolSpecialization(specialization)
        except ValueError:
            if self.address not in _warned:
                with suppress(ContractNotVerified):
                    logger.warning(
                        "ypricemagic does not recognize this pool type, please add `%s = %s` to %s.PoolSpecialization (pool=%s)",
                        await self.__build_name__,
                        specialization,
                        __name__,
                        self.address,
                    )
                _warned.add(self.address)
            return specialization

    __pool_type__: HiddenMethodDescriptor[Self, PoolSpecialization | None]

    @stuck_coro_debugger
    async def get_tvl(
        self, block: Block | None = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> UsdValue | None:
        """
        Get the total value locked (TVL) in the pool in USD.

        Args:
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            The TVL in USD, or None if it cannot be determined.

        Examples:
            >>> tvl = await pool.get_tvl()
        """
        if balances := await self.get_balances(block=block, skip_cache=skip_cache, sync=False):
            # overwrite ref to big obj with ref to little obj
            balances = iter(tuple(balances.values()))
            return UsdValue(await WeiBalance.value_usd.sum(balances, sync=False))

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def get_balances(
        self,
        block: Block | None = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: tuple[Pool, ...] = (),
    ) -> PoolBalances:
        """
        Get the balances of tokens in the pool.

        Args:
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            A dictionary mapping :class:`~ERC20` tokens to their :class:`~WeiBalance` in the pool.

        Examples:
            >>> balances = await pool.get_balances()
        """
        vault = await self.__vault__
        if vault is None:
            return {}
        tokens, balances, lastChangedBlock = await vault.get_pool_tokens(
            await self.__id__, block=block, sync=False
        )
        return {
            ERC20(token, asynchronous=self.asynchronous): WeiBalance(
                balance, token, block=block, skip_cache=skip_cache, ignore_pools=ignore_pools
            )
            for token, balance in zip(tokens, balances)
            # NOTE: some pools include themselves in their own token list, and we should ignore those
            if token != self.address
        }

    async def get_balance(
        self,
        token_address: Address,
        block: Block | None = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> WeiBalance | None:
        """
        Get the balance of a specific token in the pool.

        Args:
            token_address: The address of the token.
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            The :class:`~WeiBalance` of the specified token in the pool, or None if not found.

        Examples:
            >>> balance = await pool.get_balance("0xTokenAddress")
        """
        if info := await self.get_balances(block=block, sync=False):
            try:
                return info[token_address]
            except KeyError:
                raise TokenNotFound(token_address, self) from None

    @stuck_coro_debugger
    async def get_token_price(
        self,
        token_address: AnyAddressType,
        block: Block | None = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: tuple[Pool, ...] = (),
    ) -> UsdPrice | None:
        """
        Get the price of a specific token in the pool in USD.

        Args:
            token_address: The address of the token.
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            The price of the specified token in USD, or None if it cannot be determined.

        Examples:
            >>> price = await pool.get_token_price("0xTokenAddress")
        """
        get_balances_coro = self.get_balances(
            block=block, skip_cache=skip_cache, ignore_pools=(self, *ignore_pools), sync=False
        )
        if self.__nonweighted:
            # this await will return immediately once cached
            token_balances = await get_balances_coro
            weights = self.__weights
        else:
            token_balances, weights = await cgather(
                get_balances_coro, self.weights(block=block, sync=False)
            )
        pool_token_info = list(zip(token_balances.keys(), token_balances.values(), weights))
        for pool_token, token_balance, token_weight in pool_token_info:
            if pool_token == token_address:
                break

        paired_token_balance: WeiBalance | None = None
        for pool_token, balance, weight in pool_token_info:
            if pool_token in constants.STABLECOINS:
                paired_token_balance, paired_token_weight = balance, weight
                break
            elif pool_token == constants.WRAPPED_GAS_COIN:
                paired_token_balance, paired_token_weight = balance, weight
                break
            elif len(pool_token_info) == 2 and pool_token != token_address:
                paired_token_balance, paired_token_weight = balance, weight
                break

        if paired_token_balance is None:
            return None

        token_value_in_pool, token_balance_readable = await cgather(
            paired_token_balance.__value_usd__, token_balance.__readable__
        )
        token_value_in_pool /= paired_token_weight * token_weight
        return UsdPrice(token_value_in_pool / token_balance_readable)

    # NOTE: We can't cache this as a cached property because some balancer pool tokens can change. Womp
    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def tokens(
        self, block: Block | None = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> tuple[ERC20, ...]:
        """
        Get the tokens in the pool.

        Args:
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            A tuple of :class:`~ERC20` tokens in the pool.

        Examples:
            >>> tokens = await pool.tokens()
        """
        if self._tokens:
            return self._tokens
        tokens = tuple(
            (await self.get_balances(block=block, skip_cache=skip_cache, sync=False)).keys()
        )
        if await self.__pool_type__ in PoolSpecialization.with_immutable_tokens():
            self._tokens = tokens
        return tokens

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def weights(self, block: Block | None = None) -> list[int]:
        """
        Get the weights of tokens in the pool.

        Args:
            block: The block number to query. Defaults to the latest block.

        Returns:
            A list of weights for the tokens in the pool.

        Examples:
            >>> weights = await pool.weights()
        """
        contract = await Contract.coroutine(self.address)
        try:
            return await contract.getNormalizedWeights.coroutine(block_identifier=block)
        except AttributeError:
            # Contract has no method `getNormalizedWeights`
            self.__nonweighted = True
            num_tokens = len(await self.tokens(block=block, sync=False))
            self.__weights = [10**18 // num_tokens] * num_tokens
            return self.__weights


class BalancerV2(BalancerABC[BalancerV2Pool]):

    _pool_type = BalancerV2Pool

    _check_methods = (
        "getPoolId()(bytes32)",
        "getPausedState()((bool,uint,uint))",
        "getSwapFeePercentage()(uint)",
    )

    def __init__(self, *, asynchronous: bool = False) -> None:
        """
        Initialize a BalancerV2 instance.

        Args:
            asynchronous: Whether to use asynchronous operations.

        Examples:
            >>> balancer = BalancerV2(asynchronous=True)
        """
        super().__init__()
        self.asynchronous = asynchronous
        self.vaults = [
            BalancerV2Vault(vault, asynchronous=self.asynchronous) for vault in BALANCER_V2_VAULTS
        ]

    @stuck_coro_debugger
    async def get_token_price(
        self,
        token_address: Address,
        block: Block | None = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: tuple[Pool, ...] = (),
    ) -> UsdPrice:
        """
        Get the price of a specific token in USD.

        Args:
            token_address: The address of the token.
            block: The block number to query. Defaults to the latest block.
            skip_cache: Whether to skip the cache.

        Returns:
            The price of the specified token in USD.

        Examples:
            >>> price = await balancer.get_token_price("0xTokenAddress")
        """
        if deepest_pool := await self.deepest_pool_for(
            token_address, block=block, ignore_pools=ignore_pools, sync=False
        ):
            return await deepest_pool.get_token_price(
                token_address, block, skip_cache=skip_cache, ignore_pools=ignore_pools, sync=False
            )

    @stuck_coro_debugger
    async def deepest_pool_for(
        self,
        token_address: Address,
        block: Block | None = None,
        ignore_pools: tuple[Pool, ...] = (),
    ) -> BalancerV2Pool | None:
        """
        Find the deepest pool for a specific token.

        Args:
            token_address: The address of the token.
            block: The block number to query. Defaults to the latest block.

        Returns:
            The :class:`~BalancerV2Pool` with the deepest liquidity for the specified token, or None if not found.

        Examples:
            >>> deepest_pool = await balancer.deepest_pool_for("0xTokenAddress")
        """
        kwargs = {"token_address": token_address, "block": block}
        deepest_pools = BalancerV2Vault.deepest_pool_for.map(self.vaults, **kwargs)
        if deepest_pools := {
            vault.address: deepest_pool
            async for vault, deepest_pool in deepest_pools
            if deepest_pool is not None and deepest_pool not in ignore_pools
        }:
            logger.debug(
                "%s deepest pools for %s at %s: %s",
                self,
                token_address,
                block,
                deepest_pools,
            )
            async for pool in (
                BalancerV2Pool.get_balance.map(deepest_pools.values(), **kwargs)
                .keys(pop=True)
                .aiterbyvalues(reverse=True)
            ):
                return pool

        # TODO: afilter
        # deepest_pools = BalancerV2Vault.deepest_pool_for.map(self.vaults, **kwargs).values(pop=True).afilter()
        # async for pool in BalancerV2Pool.get_balance.map(deepest_pools, **kwargs).keys(pop=True).aiterbyvalues(reverse=True):
        #     return pool


balancer = BalancerV2(asynchronous=True)

_lookup_balance_from_tuple: Callable[[tuple[Any, T]], T] = (
    lambda pool_and_balance: pool_and_balance[1]
)
"Takes a tuple[K, V] and returns V."

_warned = set()



import importlib.util
import sys
import types
from pathlib import Path

import pytest

MODULE_PATH = Path(__file__).resolve().parents[1] / "y" / "_dank_import_guard.py"


def _load_guard_module(name: str):
    spec = importlib.util.spec_from_file_location(name, MODULE_PATH)
    module = importlib.util.module_from_spec(spec)
    loader = spec.loader
    assert loader is not None
    loader.exec_module(module)
    return module


def test_import_dank_w3_returns_underlying_object(monkeypatch):
    guard = _load_guard_module("y_dank_import_guard_success")

    sentinel = object()
    fake_dank_mids = types.ModuleType("dank_mids")
    fake_dank_mids.dank_web3 = sentinel
    monkeypatch.setitem(sys.modules, "dank_mids", fake_dank_mids)

    assert guard.import_dank_w3() is sentinel


def test_import_dank_w3_raises_actionable_error_with_cause(monkeypatch):
    guard = _load_guard_module("y_dank_import_guard_failure")

    fake_dank_mids = types.ModuleType("dank_mids")

    def _missing_attr(name: str):
        if name == "dank_web3":
            raise guard.BrowniePatchNotInitializedError("dank_web3")
        raise AttributeError(name)

    setattr(fake_dank_mids, "__getattr__", _missing_attr)
    monkeypatch.setitem(sys.modules, "dank_mids", fake_dank_mids)

    with pytest.raises(guard.DankMidsImportOrderError) as exc_info:
        guard.import_dank_w3()

    message = str(exc_info.value)
    assert "Invalid startup import/init order detected for ypricemagic" in message
    assert "Process restart is required." in message
    assert "Brownie connects before importing `y` or `dank_mids`" in message
    assert isinstance(exc_info.value.__cause__, guard.BrowniePatchNotInitializedError)



from types import SimpleNamespace

import pytest

import y._db.utils.utils as db
import y.time as ytime
from y.networks import Network


@pytest.fixture(autouse=True)
def clear_get_block_timestamp_cache():
    if hasattr(ytime.get_block_timestamp, "clear"):
        ytime.get_block_timestamp.clear(warn=False)


def test_get_block_timestamp_erigon_supports_int_like_header_timestamp(monkeypatch):
    height = 987_654_301
    expected_timestamp = 1_700_000_001
    observed = {}

    class UnixTimestampLike:
        def __int__(self):
            return expected_timestamp

    def fake_get_block_timestamp(_height, sync=False):  # noqa: ARG001
        return None

    def fake_set_block_timestamp(block, timestamp, sync=False):  # noqa: ARG001
        observed["set"] = (block, timestamp, sync)

    def fake_request_blocking(method, params):
        observed["request"] = (method, params)
        return SimpleNamespace(timestamp=UnixTimestampLike())

    monkeypatch.setattr(ytime, "CHAINID", Network.Mainnet)
    monkeypatch.setattr(ytime, "get_ethereum_client", lambda: "erigon")
    monkeypatch.setattr(db, "get_block_timestamp", fake_get_block_timestamp)
    monkeypatch.setattr(db, "set_block_timestamp", fake_set_block_timestamp)
    monkeypatch.setattr(ytime.web3.manager, "request_blocking", fake_request_blocking)

    assert ytime.get_block_timestamp(height) == expected_timestamp
    assert observed["request"] == ("erigon_getHeaderByNumber", [height])
    assert observed["set"] == (height, expected_timestamp, True)


def test_get_block_timestamp_erigon_supports_hex_timestamp_string(monkeypatch):
    height = 987_654_302
    expected_timestamp = 1_700_000_002
    observed = {}

    def fake_get_block_timestamp(_height, sync=False):  # noqa: ARG001
        return None

    def fake_set_block_timestamp(block, timestamp, sync=False):  # noqa: ARG001
        observed["set"] = (block, timestamp, sync)

    def fake_request_blocking(_method, _params):
        return SimpleNamespace(timestamp=hex(expected_timestamp))

    monkeypatch.setattr(ytime, "CHAINID", Network.Mainnet)
    monkeypatch.setattr(ytime, "get_ethereum_client", lambda: "erigon")
    monkeypatch.setattr(db, "get_block_timestamp", fake_get_block_timestamp)
    monkeypatch.setattr(db, "set_block_timestamp", fake_set_block_timestamp)
    monkeypatch.setattr(ytime.web3.manager, "request_blocking", fake_request_blocking)

    assert ytime.get_block_timestamp(height) == expected_timestamp
    assert observed["set"] == (height, expected_timestamp, True)


@pytest.mark.asyncio_cooperative
async def test_get_block_timestamp_async_erigon_supports_int_like_header_timestamp(monkeypatch):
    height = 987_654_303
    expected_timestamp = 1_700_000_003
    observed = {}

    class UnixTimestampLike:
        def __int__(self):
            return expected_timestamp

    async def fake_get_block_timestamp(_height, sync=False):  # noqa: ARG001
        return None

    def fake_set_block_timestamp(block, timestamp):
        observed["set"] = (block, timestamp)

    async def fake_get_client():
        return "erigon"

    async def fake_coro_request(method, params):
        observed["request"] = (method, params)
        return SimpleNamespace(timestamp=UnixTimestampLike())

    async def fake_fallback(_height):  # pragma: no cover
        raise AssertionError("fallback path should not run for erigon header timestamps")

    monkeypatch.setattr(ytime, "CHAINID", Network.Mainnet)
    monkeypatch.setattr(ytime, "get_ethereum_client_async", fake_get_client)
    monkeypatch.setattr(db, "get_block_timestamp", fake_get_block_timestamp)
    monkeypatch.setattr(db, "set_block_timestamp", fake_set_block_timestamp)
    monkeypatch.setattr(ytime.dank_mids.web3.manager, "coro_request", fake_coro_request)
    monkeypatch.setattr(ytime.dank_mids.eth, "get_block_timestamp", fake_fallback)

    undecorated = ytime.get_block_timestamp_async.__wrapped__.__wrapped__
    assert await undecorated(height) == expected_timestamp
    assert observed["request"] == ("erigon_getHeaderByNumber", [height])
    assert observed["set"] == (height, expected_timestamp)


@pytest.mark.asyncio_cooperative
async def test_get_block_timestamp_async_erigon_supports_hex_timestamp_string(monkeypatch):
    height = 987_654_304
    expected_timestamp = 1_700_000_004
    observed = {}

    async def fake_get_block_timestamp(_height, sync=False):  # noqa: ARG001
        return None

    def fake_set_block_timestamp(block, timestamp):
        observed["set"] = (block, timestamp)

    async def fake_get_client():
        return "erigon"

    async def fake_coro_request(_method, _params):
        return SimpleNamespace(timestamp=hex(expected_timestamp))

    monkeypatch.setattr(ytime, "CHAINID", Network.Mainnet)
    monkeypatch.setattr(ytime, "get_ethereum_client_async", fake_get_client)
    monkeypatch.setattr(db, "get_block_timestamp", fake_get_block_timestamp)
    monkeypatch.setattr(db, "set_block_timestamp", fake_set_block_timestamp)
    monkeypatch.setattr(ytime.dank_mids.web3.manager, "coro_request", fake_coro_request)

    undecorated = ytime.get_block_timestamp_async.__wrapped__.__wrapped__
    assert await undecorated(height) == expected_timestamp
    assert observed["set"] == (height, expected_timestamp)



from typing import Any, Final

from dank_mids.exceptions import BrowniePatchNotInitializedError

_INVALID_IMPORT_ORDER_MESSAGE: Final = (
    "Invalid startup import/init order detected for ypricemagic: "
    "`dank_mids` brownie patch is not initialized. "
    "Process restart is required. "
    "Fix your entrypoint startup order so Brownie connects before importing `y` or `dank_mids`."
)


class DankMidsImportOrderError(RuntimeError):
    """Raised when ypricemagic detects an invalid dank_mids/Brownie init order."""

    def __init__(self) -> None:
        super().__init__(_INVALID_IMPORT_ORDER_MESSAGE)


def import_dank_w3() -> Any:
    """Import dank_mids.dank_web3 and wrap init-order failures with actionable guidance."""
    try:
        from dank_mids import dank_web3 as dank_w3
    except BrowniePatchNotInitializedError as exc:
        raise DankMidsImportOrderError() from exc
    return dank_w3



import sys
from pathlib import Path

from setuptools import find_packages, setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

with open("requirements.txt") as f:
    requirements = [line.strip() for line in f.read().splitlines() if line.strip()]

SKIP_MYPYC = any(
    cmd in sys.argv
    for cmd in ("sdist", "egg_info", "--name", "--version", "--help", "--help-commands")
)

if SKIP_MYPYC:
    ext_modules = []
else:
    from mypyc.build import mypycify

    mypyc_args = [
        "y/_db/brownie.py",
        "y/_db/config.py",
        "y/_db/decorators.py",
        "y/_db/utils/stringify.py",
        "y/ENVIRONMENT_VARIABLES.py",
        "y/convert.py",
        "y/exceptions.py",
        "y/networks.py",
        "y/prices/utils/sense_check.py",
        "y/utils/gather.py",
        "--pretty",
        "--install-types",
        "--follow-imports=silent",
        "--disable-error-code=import-not-found",
        "--disable-error-code=no-untyped-def",
        "--disable-error-code=no-untyped-call",
    ]
    if not sys.platform.startswith("linux") or sys.maxsize < 2**32:
        # Some deps dont install properly at build time except on 64-bit Python on Linux
        # That's okay for us, we only use the [unused-ignore] code for housekeeping
        mypyc_args.append("--disable-error-code=unused-ignore")
        
    ext_modules = mypycify(mypyc_args, group_name="ypricemagic")
    


setup(
    name="ypricemagic",
    packages=find_packages(exclude=["tests", "tests.*"]),
    use_scm_version={
        "root": ".",
        "relative_to": __file__,
        "local_scheme": "no-local-version",
        "version_scheme": "python-simplified-semver",
    },
    description="Use this tool to extract historical on-chain price data from an archive node. Shoutout to @bantg and @nymmrx for their awesome work on yearn-exporter that made this library possible.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="BobTheBuidler",
    author_email="bobthebuidlerdefi@gmail.com",
    url="https://github.com/BobTheBuidler/ypricemagic",
    license="MIT",
    python_requires=">=3.10,<3.14",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: CPython",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
    ],
    setup_requires=["setuptools_scm"],
    install_requires=requirements,
    package_data={"y": ["py.typed"]},
    include_package_data=True,
    entry_points={"console_scripts": ["ypricemagic=y.cli:main"]},
    ext_modules=ext_modules,
    zip_safe=False,
)



import logging
from decimal import Decimal

import a_sync
from a_sync import cgather
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import ZERO_ADDRESS, chain
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20
from y.contracts import Contract, has_methods
from y.datatypes import Address, AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.utils.cache import a_sync_ttl_cache
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

_POOL_METHODS = (
    "poolId()(uint256)",
    "token()(address)",
    "router()(address)",
    "sharedDecimals()(uint256)",
    "localDecimals()(uint256)",
    "convertRate()(uint256)",
    "totalLiquidity()(uint256)",
    "totalSupply()(uint256)",
)

SEED_POOLS = {
    Network.Mainnet: (
        # S*USDC (USD Coin-LP)
        "0xdf0770dF86a8034b3EFEf0A1Bb3c889B8332FF56",
    )
}

_factories: dict[Address, "StargateFactory"] = {}
_factory_pools: dict[Address, set[Address]] = {}
_known_pools: set[Address] = set()
_seeds_loaded: set[int] = set()


async def _factory_from_pool(pool: Address) -> Address | None:
    router = await raw_call(
        pool,
        "router()",
        output="address",
        return_None_on_failure=True,
        sync=False,
    )
    if not router or router == ZERO_ADDRESS:
        return None
    factory = await raw_call(
        router,
        "factory()",
        output="address",
        return_None_on_failure=True,
        sync=False,
    )
    if not factory or factory == ZERO_ADDRESS:
        return None
    return factory


async def _load_factory_pools(factory: Address) -> set[Address]:
    if factory in _factory_pools:
        return _factory_pools[factory]

    stargate_factory = _factories.get(factory)
    if stargate_factory is None:
        stargate_factory = StargateFactory(factory, asynchronous=True)
        _factories[factory] = stargate_factory

    pools = await stargate_factory.pools(sync=False)
    if pools:
        _factory_pools[factory] = pools
        _known_pools.update(pools)
    return pools


async def _prime_factory_cache() -> None:
    if chain.id in _seeds_loaded:
        return
    _seeds_loaded.add(chain.id)
    for pool in SEED_POOLS.get(chain.id, ()):  # type: ignore[arg-type]
        factory = await _factory_from_pool(pool)
        if factory:
            await _load_factory_pools(factory)


@a_sync.a_sync(default="sync", cache_type="memory", ram_cache_ttl=5 * 60)
async def is_stargate_lp(token_address: AnyAddressType) -> bool:
    token_address = await convert.to_address_async(token_address)
    await _prime_factory_cache()

    if token_address in _known_pools:
        return True

    if not await has_methods(token_address, _POOL_METHODS, sync=False):
        return False

    factory = await _factory_from_pool(token_address)
    if factory:
        await _load_factory_pools(factory)

    return True


class StargateFactory(a_sync.ASyncGenericBase):
    def __init__(self, address: Address, *, asynchronous: bool = False) -> None:
        self.address = address
        self.asynchronous = asynchronous
        super().__init__()

    @a_sync.aka.cached_property
    async def contract(self) -> Contract:
        return await Contract.coroutine(self.address)

    __contract__: HiddenMethodDescriptor["StargateFactory", Contract]

    @stuck_coro_debugger
    @a_sync_ttl_cache
    async def pools(self) -> set[Address]:
        try:
            all_pools_len = await raw_call(
                self.address,
                "allPoolsLength()",
                output="int",
                return_None_on_failure=True,
                sync=False,
            )
        except Exception as exc:  # pragma: no cover - defensive, depends on node
            logger.debug("Failed to read Stargate allPoolsLength: %s", exc)
            return set()

        if not all_pools_len:
            return set()

        pool_map = a_sync.map(
            self._pool_at,
            range(all_pools_len),
            name=f"load {self} poolId",
        )
        await pool_map._init_loader
        pools = await pool_map.values(pop=True)
        return {pool for pool in pools if pool and pool != ZERO_ADDRESS}

    @stuck_coro_debugger
    async def _pool_at(self, poolid: int) -> Address | None:
        return await raw_call(
            self.address,
            "allPools(uint256)",
            inputs=poolid,
            output="address",
            return_None_on_failure=True,
            sync=False,
        )


class StargatePool(ERC20):
    @a_sync.aka.cached_property
    async def contract(self) -> Contract:
        return await Contract.coroutine(self.address)

    __contract__: HiddenMethodDescriptor["StargatePool", Contract]

    @a_sync.aka.cached_property
    async def underlying(self) -> ERC20:
        token = await raw_call(
            self.address,
            "token()",
            output="address",
            return_None_on_failure=True,
            sync=False,
        )
        if not token or token == ZERO_ADDRESS:
            raise ValueError(f"Stargate pool {self.address} has no underlying token")
        return ERC20(token, asynchronous=self.asynchronous)

    __underlying__: HiddenMethodDescriptor["StargatePool", ERC20]

    async def amount_lp_to_ld(self, amount_lp: int, block: Block | None = None) -> int | None:
        pool = await self.__contract__
        try:
            return await pool.amountLPtoLD.coroutine(amount_lp, block_identifier=block)
        except ContractLogicError:
            return None

    @stuck_coro_debugger
    async def price(
        self,
        block: Block | None = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> UsdPrice | None:
        try:
            lp_scale, underlying = await cgather(self.__scale__, self.__underlying__)
        except ValueError:
            return None

        amount_ld = await self.amount_lp_to_ld(lp_scale, block=block)
        if amount_ld is None:
            return None

        underlying_scale, underlying_price = await cgather(
            underlying.__scale__,
            underlying.price(block, skip_cache=skip_cache, sync=False),
        )
        if underlying_price is None:
            return None

        ratio = Decimal(amount_ld) / Decimal(underlying_scale)
        return UsdPrice(ratio * Decimal(underlying_price))


@a_sync.a_sync(default="sync")
async def get_price(
    token_address: AnyAddressType,
    block: Block | None = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice | None:
    pool = StargatePool(token_address, asynchronous=True)
    return await pool.price(block=block, skip_cache=skip_cache, sync=False)



# sourcery skip: merge-assign-and-aug-assign
import asyncio
import logging
import os
from json import JSONDecodeError
from random import randint
from time import time
from typing import Any, Final, final

import dank_mids
from aiohttp import ClientResponse, ClientSession, ClientTimeout, TCPConnector
from aiohttp.client_exceptions import ClientConnectorSSLError, ClientError, ContentTypeError
from async_lru import alru_cache
from dank_mids.helpers._session import HTTPStatusExtended

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import UsdPrice
from y.constants import CHAINID, NETWORK_NAME
from y.datatypes import Address, Block

logger: Final = logging.getLogger(__name__)

# current
header_env_names: Final = {"X-Signer": "YPRICEAPI_SIGNER", "X-Signature": "YPRICEAPI_SIGNATURE"}
AUTH_HEADERS: Final = {header: os.environ.get(env) for header, env in header_env_names.items()}
AUTH_HEADERS_PRESENT: Final = all(AUTH_HEADERS.values())

# some arbitrary amount of time in case the header is missing on unexpected 5xx responses
ONE_MINUTE: Final = 60
FIVE_MINUTES: Final = ONE_MINUTE * 5
TEN_MINUTES: Final = ONE_MINUTE * 10
ONE_HOUR: Final = ONE_MINUTE * 60
FALLBACK_STR: Final = "Falling back to your node for pricing."

# Five minutes is the default timeout from aiohttp.
YPRICEAPI_TIMEOUT: Final = ClientTimeout(int(os.environ.get("YPRICEAPI_TIMEOUT", FIVE_MINUTES)))

YPRICEAPI_SEMAPHORE: Final = dank_mids.BlockSemaphore(
    int(os.environ.get("YPRICEAPI_SEMAPHORE", 100))
)

if any(AUTH_HEADERS.values()) and not AUTH_HEADERS_PRESENT:
    for header in AUTH_HEADERS:
        if not AUTH_HEADERS[header]:
            raise OSError(
                f"You must also pass in a value for {header_env_names[header]} in order to use ypriceAPI."
            )


@final
class BadResponse(Exception):
    """Exception raised for bad responses from ypriceAPI."""


should_use: Final = not ENVS.SKIP_YPRICEAPI
notified: Final[set[HTTPStatusExtended]] = set()
resume_at = 0

# NOTE: if you want to bypass ypriceapi for specific tokens, have your program add the addresses to this set.
skip_tokens: Final = set()
skip_ypriceapi: Final = skip_tokens  # alias for backward compatability

#########################
# YPRICEAPI PUBLIC BETA #
#########################

_you_get: Final = (
    "access to your desired price data more quickly...",
    "...from nodes run by yearn-affiliated big brains...",
    "...on all the networks Yearn supports.",
)
_testimonials: Final = (
    "I can now get prices for all of my useless shitcoins without waiting all day for ypricemagic to load logs.",
    "I don't need to maintain an archive node anymore and that's saving me money.",
    "Wow, so fast!",
)
beta_announcement = "ypriceAPI is now in beta!\n\n"
beta_announcement += "Head to ypriceapi-beta.yearn.finance and sign up for access. You get:\n"
for you_get in _you_get:
    beta_announcement += f" - {you_get}\n"
beta_announcement += "\nCheck out some testimonials from our close frens:\n"
for testimonial in _testimonials:
    beta_announcement += f' - from anon{randint(0, 9999)}, "{testimonial}"\n'


def announce_beta() -> None:
    """Announce the beta version of ypriceAPI.

    This function logs the beta announcement message, which includes
    information about the benefits of using ypriceAPI and testimonials
    from users.

    Examples:
        >>> announce_beta()
        ypriceAPI is now in beta!
        ...
    """
    spam_your_logs_fn = logger.info if logger.isEnabledFor(logging.INFO) else print
    spam_your_logs_fn(beta_announcement)
    global should_use
    should_use = False


@alru_cache(maxsize=1)
async def get_session() -> ClientSession:
    """Get an aiohttp ClientSession for ypriceAPI.

    This session is configured with the ypriceAPI URL, headers, and timeout.

    Returns:
        An aiohttp ClientSession object.

    Examples:
        >>> session = await get_session()
    """
    return ClientSession(
        os.environ.get("YPRICEAPI_URL", "https://ypriceapi-beta.yearn.finance"),
        connector=TCPConnector(verify_ssl=False),
        headers=AUTH_HEADERS,
        timeout=YPRICEAPI_TIMEOUT,
    )


@alru_cache(ttl=TEN_MINUTES)
async def get_chains() -> dict[int, str]:
    """Get the supported chains from ypriceAPI.

    Returns:
        A dictionary mapping chain IDs to chain names.

    Examples:
        >>> chains = await get_chains()
        >>> chains[1]
        'Ethereum Mainnet'
    """
    session = await get_session()
    async with session.get("/chains") as response:
        chains = await read_response(response) or {}
    return {int(k): v for k, v in chains.items()}


@alru_cache(ttl=TEN_MINUTES)
async def chain_supported(chainid: int) -> bool:
    """Check if a chain is supported by ypriceAPI.

    Args:
        chainid: The chain ID to check.

    Returns:
        True if the chain is supported, False otherwise.

    Examples:
        >>> await chain_supported(1)
        True
        >>> await chain_supported(9999)
        False

    See Also:
        :func:`get_chains` for retrieving the list of supported chains.
    """
    if chainid in await get_chains():
        return True
    logger.info("ypriceAPI does not support %s at this time.", NETWORK_NAME)
    return False


async def get_price(token: Address, block: Block | None) -> UsdPrice | None:
    """Get the price of a token from ypriceAPI.

    Args:
        token: The address of the token.
        block: The block number to query.

    Returns:
        The price of the token in USD as an instance of :class:`UsdPrice`, or None if the price cannot be retrieved.

    This function will return None if the authorization headers are not present or if the current time is less than the resume time.

    Examples:
        >>> price = await get_price("0xTokenAddress", 12345678)
        >>> print(price)
        $1234.56780000

    Raises:
        ClientConnectorSSLError: If there is an SSL error during the request.
        asyncio.TimeoutError: If the request times out.
        ClientError: If there is a client error during the request.

    See Also:
        :class:`UsdPrice` for the price representation.
    """
    if not AUTH_HEADERS_PRESENT:
        announce_beta()
        return None

    if time() < resume_at:
        # NOTE: The reason we are here has already been logged.
        return None

    if block is None:
        block = await dank_mids.eth.block_number

    async with YPRICEAPI_SEMAPHORE[block]:
        try:
            tries = 0
            while True:
                try:
                    if not await chain_supported(CHAINID):
                        return None
                    session = await get_session()
                    async with session.get(
                        f"/get_price/{CHAINID}/{token}?block={block}"
                    ) as response:
                        return (
                            UsdPrice(price)
                            if (price := await read_response(response, token, block))
                            else None
                        )
                except ClientConnectorSSLError as e:
                    if (
                        "[[SSL: TLSV1_ALERT_INTERNAL_ERROR] tlsv1 alert internal error (_ssl.c:1129)]"
                        not in str(e)
                        or tries >= 50
                    ):
                        raise
                    tries += 1
        except asyncio.TimeoutError:
            logger.warning(f"ypriceAPI timed out for {token} at {block}.{FALLBACK_STR}")
        except ContentTypeError:
            raise
        except ClientError as e:
            logger.warning(
                f"ypriceAPI {e.__class__.__name__} for {token} at {block}.{FALLBACK_STR}"
            )


async def read_response(
    response: ClientResponse,
    token: Address | None = None,
    block: Block | None = None,
) -> Any | None:
    """Read and process the response from ypriceAPI.

    Args:
        response: The aiohttp ClientResponse object.
        token: The token address (optional).
        block: The block number (optional).

    Returns:
        The parsed response data, or None if the response is not valid.

    Raises:
        BadResponse: If the response content type is invalid.

    Examples:
        >>> response = await session.get("/get_price/1/0xTokenAddress?block=12345678")
        >>> data = await read_response(response)
    """
    # 200
    if response.status == HTTPStatusExtended.OK:
        try:
            return await response.json()
        except ContentTypeError as e:
            msg = await response.json(content_type=None)
            if msg:
                raise BadResponse(msg) from e
            logger.warning(
                f"ypriceAPI returned status code {_get_err_reason(response)} with response None. Must investigate."
            )

    # 401
    elif response.status == HTTPStatusExtended.UNAUTHORIZED:
        if HTTPStatusExtended.UNAUTHORIZED not in notified:
            logger.error(
                f"Your provided ypriceAPI credentials are not authorized for use.{FALLBACK_STR}"
            )
            notified.add(HTTPStatusExtended.UNAUTHORIZED)

    # 404
    elif response.status == HTTPStatusExtended.NOT_FOUND and token and block:
        logger.debug("Failed to get price from API: %s at %s", token, block)

    # Server Errors

    # 502 & 503
    elif response.status in {
        HTTPStatusExtended.BAD_GATEWAY,
        HTTPStatusExtended.SERVICE_UNAVAILABLE,
    }:
        logger.warning("ypriceAPI returned status code %s", _get_err_reason(response))
        try:
            try:
                msg = await response.json(content_type=None) or await response.text()
            except JSONDecodeError:
                msg = await response.text()
        except Exception:
            logger.warning(
                "exception decoding ypriceapi %s response.%s",
                response.status,
                FALLBACK_STR,
                exc_info=True,
            )
            msg = ""
                
        if msg:
            logger.warning(msg)
        _set_resume_at(_get_retry_header(response))

    else:
        msg = f"ypriceAPI returned status code {_get_err_reason(response)}"
        if token and block:
            msg += f" for {token} at {block}.{FALLBACK_STR}"
        logger.warning(msg)


def _get_err_reason(response: ClientResponse) -> str:
    """Get the error reason from a ClientResponse.

    Args:
        response: The aiohttp ClientResponse object.

    Returns:
        A string representing the error reason.

    Examples:
        >>> reason = _get_err_reason(response)
        >>> print(reason)
        [404 Not Found]
    """
    if response.reason is None:
        return f"[{response.status}]"
    ascii_encodable_reason = response.reason.encode("ascii", "backslashreplace").decode("ascii")
    return f"[{response.status} {ascii_encodable_reason}]"


def _get_retry_header(response: ClientResponse) -> int:
    retry_after: str | None = response.headers.get("Retry-After")
    return ONE_MINUTE if retry_after is None else int(retry_after)


def _set_resume_at(retry_after: float) -> None:
    """Set the resume time for ypriceAPI requests.

    Args:
        retry_after: The time to wait before retrying, in seconds.

    Examples:
        >>> _set_resume_at(60)
    """
    global resume_at
    logger.info("Falling back to your node for %s minutes.", int(retry_after / 60))
    resume_from_this_err_at = time() + retry_after
    if resume_from_this_err_at > resume_at:
        resume_at = resume_from_this_err_at
